@@ -19,148 +19,19 @@
  */
 
 #include "dialog_import_csv.h"
-#include "config.h"
+#include <libglom/libglom_config.h>
 
-#include <glom/libglom/data_structure/glomconversions.h>
+#include <libglom/data_structure/glomconversions.h>
 
 #include <gtkmm/messagedialog.h>
 #include <gtkmm/cellrenderercombo.h>
+#include <glom/utils_ui.h>
 #include <glibmm/i18n.h>
 #include <cerrno>
 
-// On Windows, "iconv" seems to be a define for "libiconv", breaking the Glib::IConv::iconv() call.
-#ifdef iconv
-#undef iconv
-#endif
-
 namespace
 {
 
-const gunichar DELIMITER = ',';
-
-#if 0
-// TODO: Perhaps we should change this to std::string to allow binary data, such
-// as images.
-// TODO: What escaping system is this? Can't we reuse some standard escaping function from somewhere? murrayc
-Glib::ustring::const_iterator advance_escape(const Glib::ustring::const_iterator& iter, const Glib::ustring::const_iterator& end, gunichar& c)
-{
-  // TODO: Throw an error if there is nothing to be escaped (iter == end)?
-  Glib::ustring::const_iterator walk = iter;
-  if(walk == end)
-    return walk;
-
-  if(*walk == 'x')
-  {
-    // Hexadecimal number, insert according unichar
-    ++ walk;
-    guint num = 0;
-
-    // TODO: Limit?
-    Glib::ustring::const_iterator pos = walk;
-    for(; (walk != end) && isxdigit(*walk); ++ walk)
-    {
-      num *= 16;
-
-      if(isdigit(*walk))
-        num += *walk - '0';
-      else if(isupper(*walk))
-        num += *walk - 'A' + 10;
-      else
-        num += *walk - 'a' + 10;
-    }
-
-    c = num;
-    return walk;
-  }
-  else if(isdigit(*walk) && *walk != '8' && *walk != '9')
-  {
-    guint num = 0;
-
-    // TODO: Limit?
-    for(; (walk != end) && isdigit(*walk) && *walk != '8' && *walk != '9'; ++ walk)
-    {
-      num *= 8;
-      num += *walk - '0';
-    }
-
-    c = num;
-    return walk;
-  }
-  else
-  {
-    switch(*walk)
-    {
-    case 'n':
-      c = '\n';
-      break;
-    case '\\':
-      c = '\\';
-      break;
-    case 'r':
-      c = '\r';
-      break;
-    case 'a':
-      c = '\a';
-      break;
-    case 'v':
-      c = '\f';
-      break;
-    case 't':
-      c = '\t';
-      break;
-    default:
-      // Unrecognized escape sequence, TODO: throw error?
-      c = *walk;
-      break;
-    }
-
-    ++ walk;
-    return walk;
-  }
-}
-#endif
-
-Glib::ustring::const_iterator advance_field(const Glib::ustring::const_iterator& iter, const Glib::ustring::const_iterator& end, Glib::ustring& field)
-{
-  Glib::ustring::const_iterator walk = iter;
-
-  gunichar quote_char = 0;
-  bool escaped = false;
-
-  field.clear();
-
-  for(; walk != end; ++ walk)
-  {
-    gunichar c = *walk;
-
-    // Skip escape sequences
-    if(escaped)
-    {
-      field += c;
-      escaped = false;
-      continue;
-    }
-
-    // Escaped stuff in quoted strings:
-    if(quote_char && c == '\\')
-      escaped = true;
-    // End of quoted string
-    else if(quote_char && c == quote_char)
-      quote_char = 0;
-    // Begin of quoted string.
-    else if(!quote_char && (c == '\'' || c == '\"'))
-      quote_char = c;
-    // End of field:
-    else if(!quote_char && c == DELIMITER)
-      break;
-
-    field += c; // Just so that we don't need to iterate through the field again, since there is no Glib::ustring::substr(iter, iter)
-  }
-
-  // TODO: Throw error if still inside a quoted string?
-  return walk;
-}
-
 struct Encoding {
   const char* name;
   const char* charset;
@@ -231,19 +102,22 @@ Glib::ustring encoding_display(const Glib::ustring& name, const Glib::ustring& c
 namespace Glom
 {
 
-Dialog_Import_CSV::Dialog_Import_CSV(BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade::Xml>& refGlade)
-: Gtk::Dialog(cobject), m_auto_detect_encoding(0), m_state(NONE)
+Dialog_Import_CSV::Dialog_Import_CSV(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>& builder)
+: Gtk::Dialog(cobject),
+  m_auto_detect_encoding()
 {
-  refGlade->get_widget("import_csv_fields", m_sample_view);
-  refGlade->get_widget("import_csv_target_table", m_target_table);
-  refGlade->get_widget("import_csv_encoding", m_encoding_combo);
-  refGlade->get_widget("import_csv_encoding_info", m_encoding_info);
-  refGlade->get_widget("import_csv_first_line_as_title", m_first_line_as_title);
-  refGlade->get_widget("import_csv_sample_rows", m_sample_rows);
-  refGlade->get_widget("import_csv_advice_label", m_advice_label);
-  refGlade->get_widget("import_csv_error_label", m_error_label);
+  builder->get_widget("import_csv_fields", m_sample_view);
+  builder->get_widget("import_csv_target_table", m_target_table);
+  builder->get_widget("import_csv_encoding", m_encoding_combo);
+  builder->get_widget("import_csv_encoding_info", m_encoding_info);
+  builder->get_widget("import_csv_first_line_as_title", m_first_line_as_title);
+  builder->get_widget("import_csv_sample_rows", m_sample_rows);
+  builder->get_widget("import_csv_advice_label", m_advice_label);
+  builder->get_widget("import_csv_error_label", m_error_label);
+#ifdef GLIBMM_EXCEPTIONS_ENABLED  
   if(!m_sample_view || !m_encoding_combo || !m_target_table || !m_encoding_info || !m_first_line_as_title || !m_sample_rows || !m_error_label)
     throw std::runtime_error("Missing widgets from glade file for Dialog_Import_CSV");
+#endif
 
   m_encoding_model = Gtk::ListStore::create(m_encoding_columns);
 
@@ -256,20 +130,29 @@ Dialog_Import_CSV::Dialog_Import_CSV(BaseObjectType* cobject, const Glib::RefPtr
   for(guint i = 0; i < N_ENCODINGS; ++ i)
   {
     iter = m_encoding_model->append();
-    if(ENCODINGS[i].name != NULL)
+    if(ENCODINGS[i].name != 0)
     {
-      (*iter)[m_encoding_columns.m_col_name] = gettext(ENCODINGS[i].name);
+      (*iter)[m_encoding_columns.m_col_name] = _(ENCODINGS[i].name);
       (*iter)[m_encoding_columns.m_col_charset] = ENCODINGS[i].charset;
     }
   }
 
+  m_sample_rows->set_value(2); //A sensible default.
+
   Gtk::CellRendererText* renderer = Gtk::manage(new Gtk::CellRendererText);
   m_encoding_combo->set_model(m_encoding_model);
   m_encoding_combo->pack_start(*renderer);
   m_encoding_combo->set_cell_data_func(*renderer, sigc::bind(sigc::mem_fun(*this, &Dialog_Import_CSV::encoding_data_func), sigc::ref(*renderer)));
   m_encoding_combo->set_row_separator_func(sigc::mem_fun(*this, &Dialog_Import_CSV::row_separator_func));
   m_encoding_combo->set_active(0);
+
   m_encoding_combo->signal_changed().connect(sigc::mem_fun(*this, &Dialog_Import_CSV::on_encoding_changed));
+
+  // TODO: Reset parser encoding on selection changed.
+  m_parser = std::auto_ptr<CsvParser>(new CsvParser(get_current_encoding().c_str()));
+  m_parser->signal_encoding_error().connect(sigc::mem_fun(*this, &Dialog_Import_CSV::on_encoding_error));
+  m_parser->signal_line_scanned().connect(sigc::mem_fun(*this, &Dialog_Import_CSV::on_line_scanned));
+
   m_first_line_as_title->signal_toggled().connect(sigc::mem_fun(*this, &Dialog_Import_CSV::on_first_line_as_title_toggled));
   m_sample_rows->signal_changed().connect(sigc::mem_fun(*this, &Dialog_Import_CSV::on_sample_rows_changed));
 
@@ -296,11 +179,26 @@ Dialog_Import_CSV::Dialog_Import_CSV(BaseObjectType* cobject, const Glib::RefPtr
   clear();
 }
 
+CsvParser::State Dialog_Import_CSV::get_parser_state() const
+{
+  return m_parser->get_state();
+}
+
+Glib::ustring Dialog_Import_CSV::get_target_table_name() const
+{
+  return m_target_table->get_text();
+}
+
+const Glib::ustring& Dialog_Import_CSV::get_filename() const
+{
+  return m_filename;
+}
+
 void Dialog_Import_CSV::import(const Glib::ustring& uri, const Glib::ustring& into_table)
 {
   clear();
 
-  Document_Glom* document = get_document();
+  Document* document = get_document();
   if(!document)
   {
     show_error_dialog(_("No Document Available"), _("You need to open a document to import the data into a table."));
@@ -321,8 +219,8 @@ void Dialog_Import_CSV::import(const Glib::ustring& uri, const Glib::ustring& in
     Gtk::TreeModel::iterator tree_iter = m_field_model->append();
     (*tree_iter)[m_field_columns.m_col_field_name] = _("<None>");
 
-    const Document_Glom::type_vecFields fields(document->get_table_fields(into_table));
-    for(Document_Glom::type_vecFields::const_iterator iter = fields.begin(); iter != fields.end(); ++ iter)
+    const Document::type_vec_fields fields(document->get_table_fields(into_table));
+    for(Document::type_vec_fields::const_iterator iter = fields.begin(); iter != fields.end(); ++ iter)
     {
       sharedptr<Field> field = *iter;
       if(!field)
@@ -351,16 +249,18 @@ void Dialog_Import_CSV::import(const Glib::ustring& uri, const Glib::ustring& in
     // Query the display name of the file to set in the title:
     m_file->query_info_async(sigc::mem_fun(*this, &Dialog_Import_CSV::on_query_info), G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME);
 
-    set_state(PARSING);
+    set_parser_state(CsvParser::STATE_PARSING);
   }
 }
 
 guint Dialog_Import_CSV::get_row_count() const
 {
-  if(m_first_line_as_title->get_active() && m_rows.size() > 1)
-    return m_rows.size() - 1;
+  const guint parser_count = m_parser->get_rows_count();
 
-  return m_rows.size();
+  if(m_first_line_as_title->get_active() && parser_count > 1)
+    return parser_count - 1;
+  else
+    return parser_count;
 }
 
 guint Dialog_Import_CSV::get_column_count() const
@@ -368,7 +268,7 @@ guint Dialog_Import_CSV::get_column_count() const
   return m_fields.size();
 }
 
-const sharedptr<Field>& Dialog_Import_CSV::get_field_for_column(guint col)
+sharedptr<const Field> Dialog_Import_CSV::get_field_for_column(guint col) const
 {
   return m_fields[col];
 }
@@ -378,7 +278,22 @@ const Glib::ustring& Dialog_Import_CSV::get_data(guint row, guint col)
   if(m_first_line_as_title->get_active())
     ++row;
 
-  return m_rows[row][col];
+  static Glib::ustring empty_result;
+
+  if(row >= m_parser->m_rows.size())
+  {
+    std::cerr << "Dialog_Import_CSV::get_data(): row out of range." << std::endl;
+    return empty_result;
+  }
+
+  const CsvParser::type_row_strings& row_data = m_parser->m_rows[row];
+  if(col >= row_data.size())
+  {
+    std::cerr << "Dialog_Import_CSV::get_data(): col out of range." << std::endl;
+    return empty_result;
+  }
+
+  return row_data[col];
 }
 
 void Dialog_Import_CSV::clear()
@@ -390,31 +305,26 @@ void Dialog_Import_CSV::clear()
   m_sample_view->set_model(m_sample_model);
   m_field_model.reset();
   m_field_model_sorted.reset();
-  m_rows.clear();
   m_fields.clear();
   m_file.reset();
-  m_stream.reset();
   m_filename.clear();
-  m_buffer.reset(NULL);
-  m_raw.clear();
-  m_parser.reset(NULL);
-
+  m_buffer.reset(0);
+  m_parser->clear();
+  //m_parser.reset(0);
   m_encoding_info->set_text("");
   m_sample_view->set_sensitive(false);
   m_encoding_combo->set_sensitive(false);
   m_first_line_as_title->set_sensitive(false);
   m_sample_rows->set_sensitive(false);
 
-  set_state(NONE);
+  set_parser_state(CsvParser::STATE_NONE);
   validate_primary_key();
 }
 
-void Dialog_Import_CSV::show_error_dialog(const Glib::ustring& primary, const Glib::ustring& secondary)
+void Dialog_Import_CSV::show_error_dialog(const Glib::ustring&, const Glib::ustring& secondary)
 {
-  Gtk::MessageDialog dialog(*this, primary, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK);
-  dialog.set_title(_("Error Importing CSV File"));
-  dialog.set_secondary_text(secondary);
-  dialog.run();
+  Utils::show_ok_dialog(_("Error Importing CSV File"),
+     secondary, *this, Gtk::MESSAGE_ERROR);
 }
 
 void Dialog_Import_CSV::encoding_data_func(const Gtk::TreeModel::iterator& iter, Gtk::CellRendererText& renderer)
@@ -432,6 +342,7 @@ bool Dialog_Import_CSV::row_separator_func(const Glib::RefPtr<Gtk::TreeModel>& /
 
 void Dialog_Import_CSV::on_query_info(const Glib::RefPtr<Gio::AsyncResult>& result)
 {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
   try
   {
     Glib::RefPtr<Gio::FileInfo> info = m_file->query_info_finish(result);
@@ -442,16 +353,28 @@ void Dialog_Import_CSV::on_query_info(const Glib::RefPtr<Gio::AsyncResult>& resu
   {
     std::cerr << "Failed to fetch display name of uri " << m_file->get_uri() << ": " << ex.what() << std::endl;
   }
+#else
+  std::auto_ptr<Glib::Error> error;
+  Glib::RefPtr<Gio::FileInfo> info = m_file->query_info_finish(result, error);
+  if (!error.get())
+  {
+    m_filename = info->get_display_name();
+    set_title(m_filename + _(" - Import From CSV File"));
+  }
+  else
+    std::cerr << "Failed to fetch display name of uri " << m_file->get_uri() << ": " << error->what() << std::endl;
+#endif    
 }
 
 void Dialog_Import_CSV::on_file_read(const Glib::RefPtr<Gio::AsyncResult>& result)
 {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
   try
   {
-    m_stream = m_file->read_finish(result);
+    m_parser->m_stream = m_file->read_finish(result);
 
     m_buffer.reset(new Buffer);
-    m_stream->read_async(m_buffer->buf, sizeof(m_buffer->buf), sigc::mem_fun(*this, &Dialog_Import_CSV::on_stream_read));
+    m_parser->m_stream->read_async(m_buffer->buf, sizeof(m_buffer->buf), sigc::mem_fun(*this, &Dialog_Import_CSV::on_stream_read));
   }
   catch(const Glib::Exception& error)
   {
@@ -459,25 +382,41 @@ void Dialog_Import_CSV::on_file_read(const Glib::RefPtr<Gio::AsyncResult>& resul
     clear();
     // TODO: Response?
   }
+#else
+    std::auto_ptr<Glib::Error> error;
+    m_parser->m_stream = m_file->read_finish(result, error);
+    if (!error.get())
+    {
+      m_buffer.reset(new Buffer);
+      m_parser->m_stream->read_async(m_buffer->buf, sizeof(m_buffer->buf), sigc::mem_fun(*this, &Dialog_Import_CSV::on_stream_read));
+    }
+    else
+    {
+      show_error_dialog(_("Could Not Open file"), Glib::ustring::compose(_("The file at \"%1\" could not be opened: %2"), m_file->get_uri(), error->what()));
+      clear();
+    }
+#endif    
 }
 
 void Dialog_Import_CSV::on_stream_read(const Glib::RefPtr<Gio::AsyncResult>& result)
 {
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
   try
   {
-    const gssize size = m_stream->read_finish(result);
-    m_raw.insert(m_raw.end(), m_buffer->buf, m_buffer->buf + size);
+    const gssize size = m_parser->m_stream->read_finish(result);
+    m_parser->m_raw.insert(m_parser->m_raw.end(), m_buffer->buf, m_buffer->buf + size);
 
     // If the parser already exists, but it is currently not parsing because it waits
     // for new input, then continue parsing.
-    if(m_parser.get() && !m_parser->idle_connection.connected())
+    // TODO: Introduce CsvParser::is_idle_handler_connected() instead?
+    if(m_parser.get() && !m_parser->m_idle_connection.connected())
     {
-      m_parser->idle_connection = Glib::signal_idle().connect(sigc::mem_fun(*this, &Dialog_Import_CSV::on_idle_parse));
+      m_parser->m_idle_connection = Glib::signal_idle().connect(sigc::mem_fun(*m_parser.get(), &CsvParser::on_idle_parse));
     }
     // If the parser does not exist yet, then create a new parser, except when the
     // current encoding does not work for the file ,in which case the user must first
     // choose another encoding.
-    else if(!m_parser.get() && m_state != ENCODING_ERROR)
+    else if(!m_parser.get() && get_parser_state() != CsvParser::STATE_ENCODING_ERROR)
     {
       begin_parse();
     }
@@ -485,13 +424,13 @@ void Dialog_Import_CSV::on_stream_read(const Glib::RefPtr<Gio::AsyncResult>& res
     if(size > 0)
     {
       // Read the next few bytes
-      m_stream->read_async(m_buffer->buf, sizeof(m_buffer->buf), sigc::mem_fun(*this, &Dialog_Import_CSV::on_stream_read));
+      m_parser->m_stream->read_async(m_buffer->buf, sizeof(m_buffer->buf), sigc::mem_fun(*this, &Dialog_Import_CSV::on_stream_read));
     }
     else
     {
       // Finished reading
-      m_buffer.reset(NULL);
-      m_stream.reset();
+      m_buffer.reset(0);
+      m_parser->m_stream.reset();
       m_file.reset();
     }
   }
@@ -501,13 +440,50 @@ void Dialog_Import_CSV::on_stream_read(const Glib::RefPtr<Gio::AsyncResult>& res
     clear();
     // TODO: Response?
   }
+#else
+    std::auto_ptr<Glib::Error> error;
+    const gssize size = m_parser->m_stream->read_finish(result, error);
+    if (!error.get())
+    {
+      m_parser->m_raw.insert(m_parser->m_raw.end(), m_buffer->buf, m_buffer->buf + size);
+
+      // If the parser already exists, but it is currently not parsing because it waits
+      // for new input, then continue parsing.
+      if(m_parser.get() && !m_parser->m_idle_connection.connected())
+      {
+        m_parser->m_idle_connection = Glib::signal_idle().connect(sigc::mem_fun(*m_parser.get(), &CsvParser::on_idle_parse));
+      }
+      // If the parser does not exist yet, then create a new parser, except when the
+      // current encoding does not work for the file ,in which case the user must first
+      // choose another encoding.
+      else if(!m_parser.get() && get_parser_state() != CsvParser::STATE_ENCODING_ERROR)
+      {
+        begin_parse();
+      }
+
+      if(size > 0)
+      {
+        // Read the next few bytes
+        m_parser->m_stream->read_async(m_buffer->buf, sizeof(m_buffer->buf), sigc::mem_fun(*this, &Dialog_Import_CSV::on_stream_read));
+      }
+      else
+      {
+        // Finished reading
+        m_buffer.reset(0);
+        m_parser->m_stream.reset();
+        m_file.reset();
+      }
+    }
+    if (error.get())
+    {
+      show_error_dialog(_("Could Not Read File"), Glib::ustring::compose(_("The file at \"%1\" could not be read: %2"), m_file->get_uri(), error->what()));
+      clear();
+    }
+#endif
 }
 
 void Dialog_Import_CSV::on_encoding_changed()
 {
-  // Reset current parsing process
-  m_parser.reset(NULL);
-
   int active = m_encoding_combo->get_active_row_number();
   switch(active)
   {
@@ -523,9 +499,15 @@ void Dialog_Import_CSV::on_encoding_changed()
     break;
   }
 
+  // Reset current parsing process
+  // TODO: Troublesome. Parser now contains a bit more members, not sure we can simply reset it like that.
+  //m_parser.reset(0);
+  //m_parser->clear();
+  //m_parser->set_encoding(get_current_encoding().c_str());
+
   // Parse from beginning with new encoding if we have already some data to
   // parse.
-  if(!m_raw.empty())
+  if(!m_parser->m_raw.empty())
     begin_parse();
 }
 
@@ -549,7 +531,7 @@ void Dialog_Import_CSV::on_first_line_as_title_toggled()
 
       // Add another row to the end, if one is loaded.
       const guint last_index = m_sample_model->children().size();
-      if(last_index < m_rows.size())
+      if(last_index < m_parser->get_rows_count())
       {
         iter = m_sample_model->append();
         (*iter)[m_sample_columns.m_col_row] = last_index;
@@ -564,7 +546,7 @@ void Dialog_Import_CSV::on_first_line_as_title_toggled()
     Gtk::TreeModel::Path path("1");
     Gtk::TreeModel::iterator iter = m_sample_model->get_iter(path);
 
-    if((!iter || (*iter)[m_sample_columns.m_col_row] != 0) && !m_rows.empty() && m_sample_rows->get_value_as_int() > 0)
+    if((!iter || (*iter)[m_sample_columns.m_col_row] != 0) && !m_parser->m_rows.empty() && m_sample_rows->get_value_as_int() > 0)
     {
       // Add first row to model
       if(!iter)
@@ -614,7 +596,8 @@ void Dialog_Import_CSV::on_sample_rows_changed()
     if(m_first_line_as_title->get_active())
       ++row_index;
 
-    for(guint i = current_sample_rows; i < new_sample_rows && row_index < m_rows.size(); ++i, ++row_index)
+    const guint rows_count = m_parser->get_rows_count();
+    for(guint i = current_sample_rows; i < new_sample_rows && row_index < rows_count; ++i, ++row_index)
     {
       Gtk::TreeModel::iterator iter = m_sample_model->append();
       (*iter)[m_sample_columns.m_col_row] = row_index;
@@ -634,6 +617,7 @@ Glib::ustring Dialog_Import_CSV::get_current_encoding() const
     return AUTODETECT_ENCODINGS[m_auto_detect_encoding].charset;
   }
 
+  // TODO: change return type?
   return encoding.c_str();
 }
 
@@ -649,29 +633,29 @@ void Dialog_Import_CSV::begin_parse()
   m_sample_model.reset();
   m_sample_view->remove_all_columns();
   m_sample_view->set_model(m_sample_model); // Empty model
-  m_rows.clear();
+  m_parser->clear();
 
-  m_parser.reset(new Parser(get_current_encoding().c_str()));
-  set_state(PARSING);
+  m_parser->set_encoding(get_current_encoding().c_str());
+  set_parser_state(CsvParser::STATE_PARSING);
 
   // Allow the Import button to be pressed when a field for the primary key
   // field is set. When the import button is pressed without the file being
   // fully loaded, the import progress waits for us to load the rest.
   validate_primary_key();
 
-  m_parser->idle_connection = Glib::signal_idle().connect(sigc::mem_fun(*this, &Dialog_Import_CSV::on_idle_parse));
+  m_parser->m_idle_connection = Glib::signal_idle().connect(sigc::mem_fun(*m_parser.get(), &CsvParser::on_idle_parse));
 }
 
-void Dialog_Import_CSV::encoding_error()
+void Dialog_Import_CSV::on_encoding_error()
 {
-  m_parser.reset(NULL);
+  m_parser->clear();
   // Clear sample preview (TODO: Let it visible, and only remove when reparsing?)
   m_sample_model.reset();
   m_sample_view->remove_all_columns();
   m_sample_view->set_model(m_sample_model); // Empty model
-  m_rows.clear();
 
-  set_state(ENCODING_ERROR);
+  // TODO: move into parser.
+  set_parser_state(CsvParser::STATE_ENCODING_ERROR);
 
   // Don't allow the import button to be pressed when an error occured. This
   // would not make sense since we cleared all the parsed row data anyway.
@@ -692,159 +676,42 @@ void Dialog_Import_CSV::encoding_error()
   }
 }
 
-bool Dialog_Import_CSV::on_idle_parse()
+/*
+ * No, this is wrong. Creating the tree model and handling a line from the CSV file are two separate steps. Proposal: Construct tree model *after* parsing, using row[0].
+ */
+void Dialog_Import_CSV::on_line_scanned(const Glib::ustring& line, guint line_number)
 {
-  // The amount of bytes to process in one pass of the idle handler
-  static const guint CONVERT_BUFFER_SIZE = 1024;
+  std::cout << "debug: on_line_scanned=" << line << std::endl;
+  if(line.empty())
+   return;
 
-  const char* inbuffer = &m_raw[m_parser->input_position];
-  char* inbuf = const_cast<char*>(inbuffer);
-  gsize inbytes = m_raw.size() - m_parser->input_position;
-  char outbuffer[CONVERT_BUFFER_SIZE];
-  char* outbuf = outbuffer;
-  gsize outbytes = CONVERT_BUFFER_SIZE;
-
-  const std::size_t result = m_parser->conv.iconv(&inbuf, &inbytes, &outbuf, &outbytes);
-  bool more_to_process = (inbytes != 0);
-
-  if(result == static_cast<size_t>(-1))
-  {
-    if(errno == EILSEQ)
-    {
-      // Invalid text in the current encoding.
-      encoding_error();
-      return false;
-    }
-
-    // If EINVAL is set, this means that an incomplete multibyte sequence was at
-    // the end of the input. We might have some more bytes, but those do not make
-    // up a whole character, so we need to wait for more input.
-    if(errno == EINVAL)
-    {
-      if(!m_stream)
-      {
-        // This means that we already reached the end of the file. The file
-        // should not end with an incomplete multibyte sequence.
-        encoding_error();
-        return false;
-      }
-      else
-      {
-        more_to_process = false;
-      }
-    }
-  }
-
-  m_parser->input_position += (inbuf - inbuffer);
-
-  // We now have outbuf - outbuffer bytes of valid UTF-8 in outbuffer.
-  const char* prev = outbuffer;
-  const char* pos;
-  const char to_find[] = { '\r', '\n', '\0' };
-
-  while( (pos = std::find_first_of<const char*>(prev, outbuf, to_find, to_find + sizeof(to_find))) != outbuf)
-  {
-    if(*pos == '\0')
-    {
-      // There is a nullbyte in the conversion. As normal text files don't
-      // contain nullbytes, this only occurs when converting for example a UTF-16
-      // file from ISO-8859-1 to UTF-8 (note that the UTF-16 file is valid ISO-8859-1,
-      // it just contains lots of nullbytes). We therefore produce an error here.
-      encoding_error();
-      return false;
-    }
-    else
-    {
-      m_parser->current_line.append(prev, pos - prev);
-      ++ m_parser->line_number;
-      if(!m_parser->current_line.empty())
-        handle_line(m_parser->current_line, m_parser->line_number);
-
-      m_parser->current_line.clear();
-
-      // Skip linebreak
-      prev = pos + 1;
-
-      // Skip DOS-style linebreak (\r\n)
-      if(*pos == '\r' && prev != outbuf && *prev == '\n') ++ prev;
-    }
-  }
-
-  // Append last chunk of this line
-  m_parser->current_line.append(prev, outbuf - prev);
-  if(!m_stream && m_raw.size() == m_parser->input_position)
-  {
-    ++ m_parser->line_number;
-    // Handle last line, if nonempty
-    if(m_parser->current_line.empty())
-      handle_line(m_parser->current_line, m_parser->line_number);
-
-    // Parsed whole file, done
-    m_parser.reset(NULL);
-    set_state(PARSED);
-  }
-
-  // Continue if there are more bytes to process
-  return more_to_process;
-}
-
-void Dialog_Import_CSV::handle_line(const Glib::ustring& line, guint line_number)
-{
-  if(line.empty()) return;
-
-  m_rows.push_back(std::vector<Glib::ustring>());
-  std::vector<Glib::ustring>& row = m_rows.back();
+  m_parser->m_rows.push_back(CsvParser::type_row_strings());
+  CsvParser::type_row_strings& row = m_parser->m_rows.back();
 
   Glib::ustring field;
   //Gtk::TreeModelColumnRecord record;
 
   // Parse first field:
-  Glib::ustring::const_iterator iter = advance_field(line.begin(), line.end(), field);
+  Glib::ustring::const_iterator line_iter = CsvParser::advance_field(line.begin(), line.end(), field);
   row.push_back(field);
 
   // Parse more fields:
-  while(iter != line.end())
+  while(line_iter != line.end())
   {
     // Skip delimiter:
-    ++ iter;
+    ++line_iter;
 
     // Read field:
-    iter = advance_field(iter, line.end(), field);
+    line_iter = CsvParser::advance_field(line_iter, line.end(), field);
 
     // Add field to current row:
     row.push_back(field);
   }
 
+  // This is the first line read if there is no model yet:
   if(!m_sample_model)
   {
-    // This is the first line read if there is no model yet:
-    m_sample_model = Gtk::ListStore::create(m_sample_columns);
-    m_sample_view->set_model(m_sample_model);
-
-    // Create field vector that contains the fields into which to import
-    // the data.
-    m_fields.resize(row.size());
-
-    Gtk::CellRendererText* text = Gtk::manage(new Gtk::CellRendererText);
-    Gtk::TreeViewColumn* col = Gtk::manage(new Gtk::TreeViewColumn(_("Line")));
-    col->pack_start(*text, false);
-    col->set_cell_data_func(*text, sigc::mem_fun(*this, &Dialog_Import_CSV::line_data_func));
-    m_sample_view->append_column(*col);
-
-    for(guint i = 0; i < row.size(); ++ i)
-    {
-      Gtk::CellRendererCombo* cell = Gtk::manage(new Gtk::CellRendererCombo);
-      Gtk::TreeViewColumn* col = Gtk::manage(new Gtk::TreeViewColumn(row[i]));
-      col->pack_start(*cell, true);
-      col->set_cell_data_func(*cell, sigc::bind(sigc::mem_fun(*this, &Dialog_Import_CSV::field_data_func), i));
-      col->set_sizing(Gtk::TREE_VIEW_COLUMN_AUTOSIZE);
-      cell->property_model() = m_field_model_sorted;
-      cell->property_text_column() = 0;
-      cell->property_has_entry() = false;
-      cell->signal_edited().connect(sigc::bind(sigc::mem_fun(*this, &Dialog_Import_CSV::on_field_edited), i));
-      m_sample_view->append_column(*col);
-    }
-
+    setup_sample_model(row);
     Gtk::TreeModel::iterator iter = m_sample_model->append();
     // -1 means the row to select target fields (see special handling in cell data funcs)
     (*iter)[m_sample_columns.m_col_row] = -1;
@@ -853,25 +720,78 @@ void Dialog_Import_CSV::handle_line(const Glib::ustring& line, guint line_number
   // Add the row to the treeview if there are not yet as much sample rows
   // as the user has chosen (note the first row is to choose the target fields,
   // not a sample row, which is why we do -1 here).
-  guint sample_rows = m_sample_model->children().size() - 1;
+  const guint sample_rows = m_sample_model->children().size() - 1;
 
   // Don't add if this is the first line and m_first_line_as_title is active:
+  const guint parser_rows_count = m_parser->get_rows_count();
   if(line_number > 1 || !m_first_line_as_title->get_active())
   {
     if(sample_rows < static_cast<guint>(m_sample_rows->get_value_as_int()))
     {
-      Gtk::TreeModel::iterator iter = m_sample_model->append();
-      (*iter)[m_sample_columns.m_col_row] = m_rows.size() - 1;
+      Gtk::TreeModel::iterator tree_iter = m_sample_model->append();
+      (*tree_iter)[m_sample_columns.m_col_row] = parser_rows_count - 1;
     }
   }
 }
 
+void Dialog_Import_CSV::setup_sample_model(CsvParser::type_row_strings& row)
+{
+  m_sample_model = Gtk::ListStore::create(m_sample_columns);
+  m_sample_view->set_model(m_sample_model);
+
+  // Create field vector that contains the fields into which to import
+  // the data.
+  //m_fields.resize(row.size());
+  m_fields.resize(row.size());
+
+  // Start with a column showing the line number.
+  Gtk::CellRendererText* text = Gtk::manage(new Gtk::CellRendererText);
+  Gtk::TreeViewColumn* col = Gtk::manage(new Gtk::TreeViewColumn(_("Line")));
+  col->pack_start(*text, false);
+  col->set_cell_data_func(*text, sigc::mem_fun(*this, &Dialog_Import_CSV::line_data_func));
+  m_sample_view->append_column(*col);
+
+  for(guint i = 0; i < row.size(); ++ i)
+  {
+    m_sample_view->append_column(*Gtk::manage(column_factory(row[i], i)));
+  }
+}
+
+Gtk::TreeViewColumn* Dialog_Import_CSV::column_factory(const Glib::ustring& title, guint index)
+{
+  Gtk::TreeViewColumn* col = new Gtk::TreeViewColumn(title);
+  Gtk::CellRendererCombo* cell = cell_factory(index);
+  col->pack_start(*Gtk::manage(cell), true);
+  col->set_cell_data_func(*cell, sigc::bind(sigc::mem_fun(*this, &Dialog_Import_CSV::field_data_func), index));
+  col->set_sizing(Gtk::TREE_VIEW_COLUMN_AUTOSIZE);
+  return col;
+}
+
+Gtk::CellRendererCombo* Dialog_Import_CSV::cell_factory(guint index)
+{
+  Gtk::CellRendererCombo* cell = new Gtk::CellRendererCombo;
+#ifdef GLIBMM_PROPERTIES_ENABLED
+  cell->property_model() = m_field_model_sorted;
+  cell->property_text_column() = 0;
+  cell->property_has_entry() = false;
+#else
+  cell->set_property("model", m_field_model_sorted);
+  cell->set_property("text-column", 0);
+  cell->set_property("has_entry", false);
+#endif
+  cell->signal_edited().connect(sigc::bind(sigc::mem_fun(*this, &Dialog_Import_CSV::on_field_edited), index));
+
+  return cell;
+}
+
 void Dialog_Import_CSV::line_data_func(Gtk::CellRenderer* renderer, const Gtk::TreeModel::iterator& iter)
 {
-  int row = (*iter)[m_sample_columns.m_col_row];
+  const int row = (*iter)[m_sample_columns.m_col_row];
   Gtk::CellRendererText* renderer_text = dynamic_cast<Gtk::CellRendererText*>(renderer);
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
   if(!renderer_text)
     throw std::logic_error("CellRenderer is not a CellRendererText in line_data_func");
+#endif
 
   if(row == -1)
     renderer_text->set_property("text", Glib::ustring(_("Target Field")));
@@ -883,66 +803,88 @@ void Dialog_Import_CSV::field_data_func(Gtk::CellRenderer* renderer, const Gtk::
 {
   const int row = (*iter)[m_sample_columns.m_col_row];
   Gtk::CellRendererCombo* renderer_combo = dynamic_cast<Gtk::CellRendererCombo*>(renderer);
+#ifdef GLIBMM_EXCEPTIONS_ENABLED
   if(!renderer_combo) throw std::logic_error("CellRenderer is not a CellRendererCombo in field_data_func");
+#endif
+
+  Glib::ustring text;
+  bool editable = false;
 
   if(row == -1)
   {
     sharedptr<Field> field = m_fields[column_number];
     if(field)
-      renderer_combo->set_property("text", field->get_name());
+      text = field->get_name();
     else
-      renderer_combo->set_property("text", Glib::ustring("<None>"));
+      text = _("<None>");
 
-    renderer_combo->set_property("editable", true);
+    editable = true;
   }
   else
   {
     // Convert to currently chosen field, if any, and back, too see how it
-    // looks like when imported
-    sharedptr<Field> field = m_fields[column_number];
-    const Glib::ustring& orig_text = m_rows[row][column_number];
+    // looks like when imported:
 
-    Glib::ustring text;
-    if(field)
+    if(column_number < m_fields.size())
     {
-      bool success;
+      sharedptr<Field> field = m_fields[column_number];
 
-      if(field->get_glom_type() != Field::TYPE_IMAGE)
+      if(row != -1 && (unsigned int)row < m_parser->get_rows_count())
       {
-        /* Exported data is always stored in postgres format */
-        const Gnome::Gda::Value value = field->from_file_format(orig_text, success);
+        const CsvParser::type_row_strings& row_strings = m_parser->m_rows[row];
+        if(column_number < row_strings.size())
+        {
+
+          const Glib::ustring& orig_text = row_strings[column_number];
+
+          if(field)
+          {
+            if(field->get_glom_type() != Field::TYPE_IMAGE)
+            {
+              /* Exported data is always stored in postgres format */
+              bool success = false;
+              const Gnome::Gda::Value value = field->from_file_format(orig_text, success);
         
-        if(!success)
-          text = _("<Import Failure>");
-        else
-          text = Glom::Conversions::get_text_for_gda_value(field->get_glom_type(), value);
-      }
-      else
-      {
-        // TODO: It is too slow to create the picture here. Maybe we should
-        // create it once and cache it. We could also think about using a
-        // GtkCellRendererPixbuf to show it, then.
-        if(!orig_text.empty() && orig_text != "NULL")
-          text = _("<Picture>");
+              if(!success)
+                text = _("<Import Failure>");
+              else
+                text = Glom::Conversions::get_text_for_gda_value(field->get_glom_type(), value);
+            }
+            else
+            {
+              // TODO: It is too slow to create the picture here. Maybe we should
+              // create it once and cache it. We could also think about using a
+              // GtkCellRendererPixbuf to show it, then.
+              if(!orig_text.empty() && orig_text != "NULL")
+                text = _("<Picture>");
+            }
+          }
+          else
+          {
+            // TODO: Should we unescape the field's content?
+            text = orig_text;
+          }
+
+          if(text.length() > 32)
+          {
+            text.erase(32);
+            text.append("…");
+          }
+
+          editable = false;
+        }
       }
     }
-    else
-    {
-      // TODO: Should we unescape the field's content?
-      text = orig_text;
-    }
-
-    if(text.length() > 32)
-    {
-      text.erase(32);
-      text.append("…");
-    }
-
-    renderer_combo->set_property("text", text);
-    renderer_combo->set_property("editable", false);
   }
+  
+  renderer_combo->set_property("text", text);
+  renderer_combo->set_property("editable", editable);
 }
 
+
+/**  Parse a row from a .cvs file. Note that this "line" might have newline 
+  *  characters inside one field value, inside quotes.
+  **/
 void Dialog_Import_CSV::on_field_edited(const Glib::ustring& path, const Glib::ustring& new_text, guint column_number)
 {
   Gtk::TreeModel::Path treepath(path);
@@ -981,18 +923,20 @@ void Dialog_Import_CSV::on_field_edited(const Glib::ustring& path, const Glib::u
   }
 }
 
-void Dialog_Import_CSV::set_state(State state)
+void Dialog_Import_CSV::set_parser_state(CsvParser::State state)
 {
-  if(m_state != state)
+  // Calling the member of a member, introduced by refactoring. TODO: clean up set_parser_state() interface.
+  if(m_parser->get_state() != state)
   {
-    m_state = state;
+    m_parser->m_state = state;
+    // Should be emitted by parser?
     m_signal_state_changed.emit();
   }
 }
 
 void Dialog_Import_CSV::validate_primary_key()
 {
-  if(m_state == NONE || m_state == ENCODING_ERROR)
+  if(get_parser_state() == (CsvParser::STATE_NONE | CsvParser::STATE_ENCODING_ERROR))
   {
     m_error_label->hide();
     set_response_sensitive(Gtk::RESPONSE_ACCEPT, false);
@@ -1010,7 +954,7 @@ void Dialog_Import_CSV::validate_primary_key()
       // and the m_fields array is not up to date. It is set in handle_line()
       // when the first line is parsed.
       primary_key_selected = false;
-      if(!m_rows.empty())
+      if(!m_parser->m_rows.empty())
       {
         for(type_vec_fields::iterator iter = m_fields.begin(); iter != m_fields.end(); ++ iter)
         {
